/*-----------------------------------------------------------------------------
 *
 *  Kaiser's Strife Script Compiler version 3
 *  A compiling utility for the Doom engine game, Strife.
 *  2/3/07 Samuel Villarreal
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * DESCRIPTION:
 *     Token handling
 *     Parse special token texts in scripts and define its actions
 *     Handle special searching routines and conversions
 *
 * Author: Kaiser
 *
 *--------------------------------------------------------------------*/

#include "localRef.h"
#include "tkParse.h"

static byte* parse;
static unsigned int lastByte;
static unsigned int tkPos = 0;
static int tkLine = 1;
static int numStates = 0;

#define MAXSTRSIZE	512

static char stringToken[32];
static int intToken;
static char dataStrToken[MAXSTRSIZE];

static unsigned short dataPos = -1;
static byte choicePos = 0;

static bool tkComment = false;

//______________________________________________________________________
//Prototypes
//______________________________________________________________________

static void getToken(void);
static void getTokenNum(void);
static void getTokenStr(void);
static void tk_FreeParse(void);
static void toUprToken(void);
static void toLwrToken(void);
static void tk_ProcessDefs(tDefTypes_t *def);
static void expectToken(const char* token, int len);
static void passUpComment(void);

//______________________________________________________________________

static tDefTypes_t npcDefTypes [] = {
	{ "DROP",			'i' },
	{ "NAME",			's' },
	{ "IF_ITEM",		'i' },
	{ "IF_ITEM2",		'i' },
	{ "IF_ITEM3",		'i' },
	{ "GOTO",			'i' },
	{ "VOICE",			's' },
	{ "PANEL",			's' },
	{ "DIALOG",			's' },
	{ NULL,				'\0' }
};
#define NUMNPCDEF (sizeof(npcDefTypes) / sizeof(tDefTypes_t))

static void tk_InitNpcDefs(void) {
	npcDefTypes[0].value = &npc[dataPos].DropType;
	npcDefTypes[1].strValue = npc[dataPos].Name;
	npcDefTypes[2].value = &npc[dataPos].ItemCheck[0];
	npcDefTypes[3].value = &npc[dataPos].ItemCheck[1];
	npcDefTypes[4].value = &npc[dataPos].ItemCheck[2];
	npcDefTypes[5].value = &npc[dataPos].Link;
	npcDefTypes[6].strValue = npc[dataPos].Sound;
	npcDefTypes[7].strValue = npc[dataPos].Backdrop;
	npcDefTypes[8].strValue = npc[dataPos].Dialogue;
}

static tDefTypes_t choiceDefTypes [] = {
	{ "GIVE",			'i' },
	{ "TEXT",			's' },
	{ "LOG",			'i' },
	{ "YES",			's' },
	{ "NO",				's' },
	{ "COST",			'x' },
	{ "NODISPLAYCOST",	'y' },
	{ "ALTCOST",		'z' },
	{ "LINK",			'i' },
	{ NULL,				'\0' }
};
#define NUMCHOICEDEF (sizeof(choiceDefTypes) / sizeof(tDefTypes_t))

static void tk_InitChoiceDefs(void) {
	choiceDefTypes[0].value = &npc[dataPos].Responses[choicePos].GiveType;
	choiceDefTypes[1].strValue = npc[dataPos].Responses[choicePos].Reply;
	choiceDefTypes[2].value = &npc[dataPos].Responses[choicePos].Log;
	choiceDefTypes[3].strValue = npc[dataPos].Responses[choicePos].Yes;
	choiceDefTypes[4].strValue = npc[dataPos].Responses[choicePos].No;
	choiceDefTypes[5].value = &npc[dataPos].Responses[choicePos].Item[0];
	choiceDefTypes[6].value = &npc[dataPos].Responses[choicePos].Item[1];
	choiceDefTypes[7].value = &npc[dataPos].Responses[choicePos].Item[2];
	choiceDefTypes[8].value = &npc[dataPos].Responses[choicePos].Link;
}

//______________________________________________________________________
//tk_ParseScript: Scan though until it hits the last byte in the file.
//Scanning will begin once ACTOR is found and stops after END is found.
//______________________________________________________________________

void tk_ParseScript(const byte* stream, const int size) {
	char text[32];
	bool foundToken = false;
	parse = new byte[size];
	memcpy(parse, stream, size);
	lastByte = size;

	tkLine = 1;

	while( tkPos != lastByte ) {
		expectToken("ACTOR", 8);
		getTokenNum();	//get the conversation ID
		kput("\nReading Block ID %i", intToken);

		choicePos = 0;
		dataPos++;
		npc[dataPos].SpeakerType = intToken;
		tk_InitNpcDefs();	//associate the next NPC array with the deftype.

		while( strncmp(stringToken, "END", 4) ) {
			foundToken = false;
			getToken();
			toUprToken();

			for( int i = 0; i < NUMNPCDEF; i++ ) {
				snprintf(text, 32, "%s", npcDefTypes[i].field);
				if(!strncmp(stringToken, text, 32)) {
					foundToken = true;
					tk_ProcessDefs(&npcDefTypes[i]);
					break;
				}
			}

			/*if(foundToken == false)
				kerr("Line %i: Unknown token %s", tkLine, stringToken);*/

			/*When choice tokens are found, it will scan though for choice tokens
			until ';' is specified to stop checking for choice options*/
			if(!strncmp(stringToken, "CHOICE", 8)) {

				tk_InitChoiceDefs();
				kput("Reading Choice# %i", choicePos);

				while( strncmp(stringToken, ";", 2) ) {
					getToken();
					toUprToken();

					for( int i = 0; i < NUMCHOICEDEF; i++ ) {
						snprintf(text, 32, "%s", choiceDefTypes[i].field);
						if(!strncmp(stringToken, text, 32)) {
							tk_ProcessDefs(&choiceDefTypes[i]);
							break;
						}
					}

					if(!strncmp(stringToken, "CHOICE", 8))
						kerr("Line %i: Choice has been defined inside an unclosed block\nMake sure the block is closed with ';'", tkLine);
				}

				/*If no text is specified for the NO and YES options, then assign
				the default text to them as well as attach the text in the REPLY field*/
				if(!npc[dataPos].Responses[choicePos].No[0])
					snprintf(npc[dataPos].Responses[choicePos].No, 80,
						"%s%s", DEFAULTNO, npc[dataPos].Responses[choicePos].Reply);

				if(!npc[dataPos].Responses[choicePos].Yes[0])
					snprintf(npc[dataPos].Responses[choicePos].Yes, 80,
						"%s%s", DEFAULTYES, npc[dataPos].Responses[choicePos].Reply);

				choicePos++;
			}

			if(!strncmp(stringToken, "ACTOR", 8))
				kerr("Line %i: Actor has been defined inside an unclosed block\nMake sure the block is closed with 'END'", tkLine);
		}
	}

	tk_FreeParse();
}

//______________________________________________________________________
//tk_ProcessDefs: Process each char type based on what the token is used for
//x, y and z are used only for the COST field, basically used to associate
//the amount values when the token is found.
//______________________________________________________________________

static void tk_ProcessDefs(tDefTypes_t *def) {
	switch(def->type) {
		case 'i':	//find an integer value
			getTokenNum();
			*def->value = intToken;
			break;
		case 's':	//find a quoted string
			getTokenStr();
			memcpy(def->strValue, dataStrToken, MAXSTRSIZE);
			break;
		case 'x':	//find an integer value and then find the amount
			getTokenNum();
			*def->value = intToken;

			expectToken("*", 2);
			getTokenNum();
			npc[dataPos].Responses[choicePos].Count[0] = intToken;
			break;
		case 'y':	//find an integer value and then find the amount
			getTokenNum();
			*def->value = intToken;

			expectToken("*", 2);
			getTokenNum();
			npc[dataPos].Responses[choicePos].Count[1] = intToken;
			break;
		case 'z':	//find an integer value and then find the amount
			getTokenNum();
			*def->value = intToken;

			expectToken("*", 2);
			getTokenNum();
			npc[dataPos].Responses[choicePos].Count[2] = intToken;
			break;
		default:
			kerr("tk_ProcessDefs: Unknown type: %c", def->type);
			break;
	}
}

//______________________________________________________________________
//expectToken: Forces a token to be checked and not overlooked
//______________________________________________________________________

static void expectToken(const char* token, const int len) {
	getToken();
	toUprToken();

	if(strncmp(stringToken, token, len))
		kerr("Line %i: Expected %s, but found %s", tkLine, token, stringToken);
}

//______________________________________________________________________
//getToken: Basic routine search function. Will easily overlook any non-valid
//tokens. Basically just collects words together.
//______________________________________________________________________

static void getToken(void) {
	bool foundToken = false;
	int i = 0;
	memset(&stringToken, 0, 32);
	do {
			if(parse[tkPos] == '\n') tkLine++;
			if(parse[tkPos] > 0x20) {
				foundToken = true;
				stringToken[i++] = parse[tkPos];
			}
			else if(foundToken == true) {
				return;
			}
	}	while(++tkPos != lastByte);
}

//______________________________________________________________________
//getTokenNum:	Search for any valid number string.
//______________________________________________________________________

static void getTokenNum(void) {
	bool foundToken = false;
	int i = 0, k;
	char text[32];
	intToken = 0;
	do {
			if(parse[tkPos] == '\n') tkLine++;
			if(parse[tkPos] == 45 || (parse[tkPos] >= 48 && parse[tkPos] <= 57)) {
				foundToken = true;
				text[i++] = parse[tkPos];
			}
			else {
				if(foundToken == true) {
					int pow = 1;
					for(k = i-1; k >= 0; k--) {
						(text[k] == 45) ? intToken *= -1 :
						(intToken += (text[k] - 48) * pow);
						pow *= 10;
					}
					return;
				}
				else {
					if(parse[tkPos] > 0x20) {
						kerr("getTokenNum Line %i: Expected number after %s",
							tkLine, stringToken);
					}
				}
			}
	}	while(tkPos++ != lastByte);
}

//______________________________________________________________________
//getTokenStr:	Gather strings specified in quotes
//______________________________________________________________________

static void getTokenStr(void) {
	int i = 0;
	memset(&dataStrToken, 0, MAXSTRSIZE);
	do {
			if(parse[tkPos] == '\n') tkLine++;
			if(parse[tkPos] == '"') {
				tkPos++;
				do {
					if(parse[tkPos] == '\n') {
						kerr("getTokenStr Line %i: Newline in string..\n", tkLine);
					}
					if(tkPos >= lastByte) {
						kerr("getTokenStr Line %i: End of file before completing string..\n", tkLine);
					}
					dataStrToken[i++] = parse[tkPos];
				} while(parse[++tkPos] != '"');
				return;
			}
			else if(!(parse[tkPos] <= 0x20)) {
				kerr("getTokenStr Line %i: String does not begin with quotes..");
			}
		} while(tkPos++ != lastByte);
}

//______________________________________________________________________
//getTokenLen
//______________________________________________________________________

static int getTokenLen(void) {
	int i;
	for(i=0;;i++) {
		if(stringToken[i] == 0) break;
	}
	return i;
}

//______________________________________________________________________
//toUprToken
//______________________________________________________________________

static void toUprToken(void) {
	int i;
	for(i = 0; i < getTokenLen(); i++) {
		if(stringToken[i] >= 'a' && stringToken[i] <= 'z') {
			stringToken[i] -= 32;
		}
	}
}

//______________________________________________________________________
//toLwrToken
//______________________________________________________________________

static void toLwrToken(void) {
	int i;
	for(i = 0; i < getTokenLen(); i++) {
		if(stringToken[i] >= 'A' && stringToken[i] <= 'Z') {
			stringToken[i] += 32;
		}
	}
}

//______________________________________________________________________
//tk_FreeParse
//______________________________________________________________________

static void tk_FreeParse(void) {
	delete [] parse;
	tkLine = 1;
	tkPos = 0;
	dataPos = -1;
}

/*tkParse.cpp eof*/